
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : init-images.scm
;; DESCRIPTION : setup image converters
;; COPYRIGHT   : (C) 2003  Joris van der Hoeven
;;
;; This software falls under the GNU general public license version 3 or later.
;; It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
;; in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (convert images init-images))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Helper functions for testing available converters
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define inkscape-check -1)
(define gs-check -1)
(define rsvg-check -1)

(define (imagemagick-binary)
  (if (os-mingw?) "magick" ;; avoid collision with Windows native command (imagemagick v>=7.0)
       "convert"))

(tm-define (has-convert?)
  (url-exists-in-path? (imagemagick-binary)))

(tm-define (has-pdftocairo?)
  (url-exists-in-path? "pdftocairo"))

(tm-define (has-inkscape?)
  (if (== inkscape-check -1)
          (with ans (and (url-exists-in-path? "inkscape")
                     (test-version-num "inkscape -V" "Inkscape " 0 92))
             (set! inkscape-check (if ans 1 0))
             ans)
        (> inkscape-check 0)))


(tm-define (has-rsvg-convert?)
  (if (== rsvg-check -1)
          (with ans (and (url-exists-in-path? "rsvg-convert")
                    (test-version-num "rsvg-convert -v" "version " 2 40))
             (set! rsvg-check (if ans 1 0))
             ans)
        (> rsvg-check 0)))

(define (gs-url)
   (let* ((n1 "$TEXMACS_PATH\\bin\\gs.exe;c:\\Program F*\\gs\\gs*\\gswin*c.exe")
         (n2 "$TEXMACS_PATH/bin/gs:gs")
         (name (if (os-mingw?) n1 n2)))
         (url->system (url-resolve-in-path name))))

(define (gs-binary)
    (escape-shell (gs-url)))

(tm-define (has-gs?)
  (if (== gs-check -1)
        (with ans (and (url-exists? (gs-url))
                  (test-version-num (string-append (gs-binary) " --version") "" 9 26))
; because of bug #61991, but maybe a good idea to require even more recent gs (9.50?)
; see https://www.cvedetails.com/product/36469/Artifex-Ghostscript.html?vendor_id=10846
          (set! gs-check (if ans 1 0))
          (if ans #t
           (begin
              (debug-message "debug-convert" "No Ghostscript for Scheme converters\n")
            #f)))
      (> gs-check 0)))

(define (test-version-num cmd key major minor)
;tests if program reports version >=  major.minor
  (with a (eval-system cmd)
    (if (string-null? a)
      #f
      (with st (if (string-null? key)
               0 (string-search-forwards key 0 a))
        (if (== st -1)
          #f
          (let*
            ((a (substring a (+ st (string-length key))))
             (a (car (string-split a #\space)))
             (a (string-split a #\.))
             (v1 (string->number (first a)))
             (v2 (string->number (second a))))
            (if (or (> v1 major) (and (== v1 major) (>= v2 minor)))
              #t
              (begin
                (debug-message "debug-convert" (string-append cmd " reports version " (first a) "." (second a)
                 ";\nTeXmacs will not use it as a converter.\n
                 (minimun required version is " (number->string major) "." (number->string minor) ")\n"))
                #f))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Helpers and function-with-option(s) for converters
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-preferences
  ("texmacs->image:raster-resolution" "300" noop))

(define (get-raster-resolution opts)
  (or (assoc-ref opts "texmacs->image:raster-resolution")
      (get-preference "texmacs->image:raster-resolution")))

(define (get-shell-command cmd)
  (if (not (os-mingw?)) cmd
      (escape-shell (url-concretize (url-resolve-in-path cmd)))))

(tm-define (rsvg-convert x opts)
  ;(display* "scm rsvg-convert " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (fm (format-default-suffix (assoc-ref opts 'to-format)))
         (fm (if (== fm "ps") "eps" fm))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append "." fm))))
         (res (get-raster-resolution opts))
         (w (assoc-ref opts 'width)) ; width for export
         (wstr (if w (string-append " -w "  w) ""))
         (h (assoc-ref opts 'height))
         (hstr (if h (string-append " -h "  h) ""))
         (dstr (if (or (in? fm '("eps" "pdf")) h w) "" (string-append " -d " res)))
         (cmd (string-append (get-shell-command "rsvg-convert")
               " -f " fm wstr hstr dstr " -o ")))
    ;(display* cmd dest " " x "\n")
    (system-2 cmd dest x)
    (if (url-exists? dest) dest #f)))

(tm-define (inkscape-convert x opts)
  ;(display* "scm inkscape-convert " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (fm (format-default-suffix (assoc-ref opts 'to-format)))
         (fm (if (== fm "ps") "eps" fm))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append "." fm))))
         (fullname (url-concretize dest))
         (res (get-raster-resolution opts))
         (w (assoc-ref opts 'width)) ; width in pixels for raster export
         (wstr (if w (string-append " -w "  w) ""))
         (h (assoc-ref opts 'height))
         (hstr (if h (string-append " -h "  h) ""))
         (dstr (if (or h w) "" (string-append " -d " res)))
         (cmd (string-append (get-shell-command "inkscape")
               " --export-overwrite --pdf-poppler -l --export-type=" fm
               hstr wstr dstr " -o ")))
    ;(display* cmd dest " " x "\n")
    (system-2 cmd dest x)
    (if (url-exists? dest) dest #f)))

(tm-define (pdf-file->pdftocairo-raster x opts)
  ;(display* "scm pdf-file->pdftocairo-raster " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (fm (format-default-suffix (assoc-ref opts 'to-format)))
         (fm (if (== fm "ps") "eps" fm))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append "." fm))))
         (fullname (url-concretize dest))
         (suffix (url-suffix fullname))
         (name (string-drop-right fullname (+ 1 (string-length suffix))))
         (transp (if (== fm "png") "-transp " ""))
         (w (assoc-ref opts 'width)) ; width in pixels for raster export
         (wstr (if w (string-append " -scale-to-x "  w) ""))
         (h (assoc-ref opts 'height))
         (hstr (if h (string-append " -scale-to-y "  h) ""))
         (res (get-raster-resolution opts))
         (dstr (if (or h w) "" (string-append " -r " res)))
         (cmd (string-append (get-shell-command "pdftocairo")" -singlefile " transp "-" fm
               hstr wstr dstr)))
    ;(display cmd " " x " "  name)
    (system-2 cmd x name)
    (if (url-exists? dest) dest #f)))

(tm-define (imagemagick-any->raster x opts)
   ;(display* "scm imagemagick-any->raster " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (fm (format-default-suffix (assoc-ref opts 'to-format)))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append "." fm))))
         (orig (url-concretize x))
         (w (assoc-ref opts 'width)) ; width in pixels
         (h (assoc-ref opts 'height))
         (geom (if (or w h) (string-append " -resize "  w "x" h "!") ""))
         (res (get-raster-resolution opts)))
    ;(display (string-append (get-shell-command (imagemagick-binary)) " -density " res " " orig " " geom " "  dest "\n"))
    (system-1 (string-append (get-shell-command (imagemagick-binary)) " -density " res " " orig " " geom) dest)
    (if (url-exists? dest) dest #f)))

(tm-define (gs-pdf2eps x opts)
  ;(display* "scm gs-pdf2eps " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append ".eps")))))
        (cpp-gspdf2eps x dest)
  ;; the converter in c++ properly keeps the bounding box, unlike the previous code
    (if (url-exists? dest) dest #f))
  )

(tm-define (gs-ps2pdf x opts)
  (display* "scm gs-ps2pdf " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append ".pdf"))))
         (pdfv (get-preference "texmacs->pdf:version"))
         (pdfv (if (== pdfv "default") "1.4" pdfv))
         (cmd (string-append (gs-binary) " -dQUIET -dNOPAUSE -dBATCH -dSAFER -sDEVICE=pdfwrite "
                " -dUseCropBox -dFitPage -dAutoRotatePages=/None "
                " -dCompatibilityLevel=" pdfv
                " -sOutputFile=" (url-sys-concretize dest)
                " -f " (url-sys-concretize x))))
    (display* cmd "\n")
    (display* (eval-system cmd) "\n")
    (if (url-exists? dest) dest #f))
  )

(tm-define (pdf-file->gs-raster x opts)
  ;(display* "scm pdf-file->gs-raster " x opts "\n")
  (let* ((last? (assoc-ref opts 'last?))
         (fm (format-default-suffix (assoc-ref opts 'to-format)))
         (dev (cond ((== fm "jpg") "jpeg")
                    ((== fm "tif") "tiff24nc")
                    (else "pngalpha")))
         (dest (if last? (assoc-ref opts 'dest)
               (url-glue (url-temp) (string-append "." fm))))
         (res (get-raster-resolution opts))
         (w (assoc-ref opts 'width)) ; width in pixels
         (h (assoc-ref opts 'height))
         (gs (gs-binary))
         (cl (list gs " -dBATCH" " -dNOPAUSE" " -dSAFER"
                           " -dNOPROMPT" (string-append " -sDEVICE=" dev) " -dUseCropBox"))
         (cl (if (and w h) (append cl `(" -dPDFFitPage" ,(string-append " -g" w "x" h)) )
                           (rcons cl (string-append " -r" res))  ))
         (cl (append cl `(,(string-append " -sOutputFile=" (url-sys-concretize dest))
                           ,(string-append " " (url-sys-concretize x)))))
         (cmd (apply string-append cl)))
    ;(display* cl "\n")
    ;(evaluate-system cl '() '() '(1 2)) ;for some reason this gives non-zero return code in Windows
    (with ans (eval-system cmd)
      (if (url-exists? dest) dest
        (begin (debug-message "debug-convert" (string-append "pdf-file->gs-raster failed\n"
                 "command: " cmd "\n"
                 "produced output: " ans "\n"))
           #f)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Formats and converters
;; Note : Multiple converters can be defined for a given conversion
;; (e.g. pdftocairo and pdf2svg for pdf -> svg). In that case,
;; TeXmacs will use the latest defined converter that is available on the system
;; => converters defined in my-init-texmacs.scm override similar ones defined here.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Graphical document and geometric image formats
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-format postscript
  (:name "Postscript")
  (:suffix "ps" "eps"))

(define-format pdf
  (:name "Pdf")
  (:suffix "pdf"))

(converter pdf-file postscript-file
  (:require (has-gs?))
  (:function-with-options gs-pdf2eps))

(converter postscript-file pdf-file
  (:require (has-gs?))
  (:function-with-options gs-ps2pdf))

(define-format xfig
  (:name "Xfig")
  (:suffix "fig"))

;(converter xfig-file postscript-file
;  (:shell "fig2ps" from to))
  
(converter xfig-file pdf-file
  (:require (url-exists-in-path? "fig2pdf"))
  (:shell "fig2pdf" from to))

; as of 2023 grace is still available in linux distros but it seems a dead project
; there are now several other OSS Wysiwyg graphers
;(define-format xmgrace
;  (:name "Xmgrace")
;  (:suffix "agr" "xmgr"))

;(converter xmgrace-file postscript-file
;  (:require (url-exists-in-path? "xmgrace"))
;  (:shell "xmgrace" "-noask -hardcopy -hdevice EPS -printfile" to from))

(define-format svg
   (:name "Svg")
   (:suffix "svg"))

(converter svg-file pdf-file
  (:require (has-rsvg-convert?))
  (:function-with-options rsvg-convert))

;(converter svg-file postscript-file
;  (:require (has-rsvg-convert?))
;  (:function-with-options rsvg-convert))

;(converter svg-file postscript-file
;  (:require (has-inkscape?))
;  (:function-with-options inkscape-convert))

(converter svg-file pdf-file
  (:require (has-inkscape?))
  (:function-with-options inkscape-convert))

(define-format geogebra
  (:name "Geogebra")
  (:suffix "ggb"))

(converter geogebra-file pdf-file
  (:require (url-exists-in-path? "geogebra"))
  (:shell "geogebra" "--export=" to from))

(converter geogebra-file png-file
  (:require (url-exists-in-path? "geogebra"))
  (:shell "geogebra" "--export=" to "--dpi=600" from))

(converter pdf-file svg-file
  (:require (has-inkscape?))
  (:function-with-options inkscape-convert))

(converter pdf-file svg-file
  (:require (url-exists-in-path? "pdf2svg"))
  (:shell "pdf2svg" from to))

(converter pdf-file svg-file
  (:require (has-pdftocairo?))
  (:shell "pdftocairo" "-origpagesizes -nocrop -nocenter -svg" from to))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Bitmap image formats
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-format xpm
  (:name "Xpm")
  (:suffix "xpm"))

(converter xpm-file ppm-file
  (:require (has-convert?))
  (:shell "convert" from to))

(define-format jpeg
  (:name "Jpeg")
  (:suffix "jpg" "jpeg"))

(converter jpeg-file postscript-document
  (:function image->psdoc))

(converter jpeg-file pnm-file
  (:require (has-convert?))
  (:shell "convert" from to))

(define-format tif
  (:name "Tif")
  (:suffix "tif" "tiff"))

;(converter tif-file postscript-document
;  (:function image->psdoc))
  
(converter tif-file png-file
  (:require (has-convert?))
  (:shell "convert" from to))
  
(converter tif-file pdf-file
  (:require (has-convert?))
  (:shell "convert" from to))

(define-format ppm
  (:name "Ppm")
  (:suffix "ppm"))

(converter ppm-file gif-file
  (:require (has-convert?))
  (:shell "convert" from to))

(define-format gif
  (:name "Gif")
  (:suffix "gif"))

(converter gif-file postscript-document
  (:function image->psdoc))

(converter gif-file pnm-file
  (:require (has-convert?))
  (:shell "convert" from to))

(define-format png
  (:name "Png")
  (:suffix "png"))

;(converter png-file postscript-document
;  (:function image->psdoc))

;(converter png-file pnm-file
;  (:require (has-convert?))
;  (:shell "convert" from to))

(converter geogebra-file png-file
  (:require (url-exists-in-path? "geogebra"))
  (:shell "geogebra" "--export=" to "--dpi=600" from))

(define-format pnm
  (:name "Pnm")
  (:suffix "pnm"))

;(converter pnm-file postscript-document
;  (:function image->psdoc))

(converter pnm-file png-file
  (:require (has-convert?))
  (:shell "convert" from to))

(converter pnm-file png-file
  (:require (has-convert?))
  (:shell "convert" from to))

(define-format webp
  (:name "Webp")
  (:suffix "webp"))

(converter webp-file png-file
  (:require (url-exists-in-path? "dwebp"))
  (:shell "dwebp" from "-o" to))

(converter png-file webp-file
  (:require (url-exists-in-path? "cwebp"))
  (:shell "cwebp" from "-o" to))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; From vector graphics to bitmaps
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(converter pdf-file png-file
  (:require (has-convert?))
  (:function-with-options imagemagick-any->raster))
  
(converter pdf-file jpeg-file
  (:require (has-convert?))
  (:function-with-options imagemagick-any->raster))
 
(converter pdf-file png-file
  (:require (has-pdftocairo?))
  (:function-with-options pdf-file->pdftocairo-raster)
  ;;(:option "texmacs->image:raster-resolution" "450")
  ;;if this is set it overrides the preference widget settings
  )

(converter pdf-file jpeg-file
  (:require (has-pdftocairo?))
  (:function-with-options pdf-file->pdftocairo-raster))

(converter pdf-file png-file
  (:require (has-gs?))
  (:function-with-options pdf-file->gs-raster))
  
(converter pdf-file jpeg-file
  (:require (has-gs?))
  (:function-with-options pdf-file->gs-raster))
 
(converter svg-file png-file
  (:require (has-rsvg-convert?))
  (:function-with-options rsvg-convert))

(converter svg-file png-file
  (:require (has-inkscape?))
  (:function-with-options inkscape-convert))
